//
//  ItemVC.swift
//  Groceries
//
//  Created by Tim Roadley on 4/10/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import Foundation
import CoreData

class ItemVC: GenericVC, UITextFieldDelegate, CDPickerTextFieldDelegate, UIImagePickerControllerDelegate, UINavigationControllerDelegate {
    
    @IBOutlet var scrollView: UIScrollView!
    @IBOutlet var nameTextField: UITextField!
    @IBOutlet var quantityTextField: UITextField!
    @IBOutlet var unitPickerTextField: UnitPickerTF!
    @IBOutlet var homeLocationPickerTextField: LocationAtHomePickerTF!
    @IBOutlet var shopLocationPickerTextField: LocationAtShopPickerTF!
    @IBOutlet var photoImageView: UIImageView!
    @IBOutlet var cameraButton: UIButton!
    var camera:UIImagePickerController?
    
    // MARK: - DELEGATE: UITextField
    func textFieldDidBeginEditing(textField: UITextField) {
        
        self.activeField = textField
        
        switch (textField) {
            case self.nameTextField:
                if self.nameTextField.text == "New Item" {
                    self.nameTextField.text = ""
                }
                break;
            case self.unitPickerTextField:
                self.unitPickerTextField.performFetch()
                self.unitPickerTextField.setSelectedRowForComponent(0)
                break;
            case self.homeLocationPickerTextField:
                self.homeLocationPickerTextField.performFetch()
                self.homeLocationPickerTextField.setSelectedRowForComponent(0)
                break;
            case self.shopLocationPickerTextField:
                self.shopLocationPickerTextField.performFetch()
                self.shopLocationPickerTextField.setSelectedRowForComponent(0)
                break; 
            default:break;
        }
    }
    func textFieldDidEndEditing(textField: UITextField) {
     
        self.activeField = nil
        
        if let item = self.selectedObject as? Item {

            switch (textField) {
                case self.nameTextField:
                     if self.nameTextField.text == "" {
                        self.nameTextField.text = "New Item"
                     }
                     item.name = self.nameTextField.text
                break;
                case self.quantityTextField:
                     if let text = self.quantityTextField.text, floatValue = Float(text)  {    
                         item.quantity = NSNumber(float: floatValue)
                     }
                break;
                default:break;
            }
        } else {self.done()}
    }
    
    // MARK: - VIEW
    func refreshInterface () {
        if let item = self.selectedObject as? Item {
            
            item.modified = NSDate()
            
            self.nameTextField.text = item.name
            self.quantityTextField.text = item.quantity?.stringValue        
            self.unitPickerTextField.text = item.unit?.name ?? ""
            self.unitPickerTextField.selectedTitle = item.unit?.name ?? ""
            self.homeLocationPickerTextField.text = item.locationAtHome?.storedIn ?? ""
            self.homeLocationPickerTextField.selectedTitle = item.locationAtHome?.storedIn ?? ""
            self.shopLocationPickerTextField.text = item.locationAtShop?.aisle ?? ""
            self.shopLocationPickerTextField.selectedTitle = item.locationAtShop?.aisle ?? ""
            
            if let photoData = item.photo?.data, let image = UIImage(data: photoData) {            
                CDHelper.shared.context.performBlock {
                    self.photoImageView.image = image
                }
            }        
            self.checkCamera()        
        } else {self.done()}
    }
    override func viewDidLoad() {
        super.viewDidLoad()
        self.hideKeyboardWhenBackgroundIsTapped()
        self.nameTextField.delegate = self
        self.quantityTextField.delegate = self
        let doneButton = UIBarButtonItem(barButtonSystemItem: .Done, target: self, action: "done")
        self.navigationItem.rightBarButtonItem = doneButton
                
        self.unitPickerTextField.delegate = self
        self.unitPickerTextField.pickerDelegate = self
        self.homeLocationPickerTextField.delegate = self
        self.homeLocationPickerTextField.pickerDelegate = self
        self.shopLocationPickerTextField.delegate = self
        self.shopLocationPickerTextField.pickerDelegate = self
    }
    override func viewWillAppear(animated: Bool) {
        super.viewWillAppear(animated)
            
        // Register for keyboard notifications while the view is visible.
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "keyboardDidShow:", name: UIKeyboardDidShowNotification, object: self.view.window)
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "keyboardWillHide", name: UIKeyboardWillHideNotification, object: self.view.window)
            
        self.validateItem()
        self.refreshInterface()
        if self.nameTextField.text == "New Item" {
            self.nameTextField.text = ""
            self.nameTextField.becomeFirstResponder()
        }
    }
    override func viewDidDisappear(animated: Bool) {
        super.viewDidDisappear(animated)
        self.validateItem()
        CDHelper.saveSharedContext()
        
        // Unregister for keyboard notifications while the view is not visible.
        NSNotificationCenter.defaultCenter().removeObserver(self, name: UIKeyboardDidShowNotification, object: self.view.window)
        NSNotificationCenter.defaultCenter().removeObserver(self, name: UIKeyboardWillHideNotification, object: self.view.window)
        
        // Turn item & item photo into a fault
        if let item = self.selectedObject as? Item {
            if let itemPhoto = item.photo {
                CDHelper.shared.context.refreshObject(itemPhoto, mergeChanges: false)
            }
            CDHelper.shared.context.refreshObject(item, mergeChanges: false)
        }
    }
    
    // MARK: - DATA
    func validateItem () {
        if let item = self.selectedObject as? Item {
            Item.ensureHomeLocationIsNotNil(item)
            Item.ensureShopLocationIsNotNil(item)
        } else {self.done()}
    }
    
    // MARK: - DELEGATE: CDPickerTextFieldDelegate
    func selectedObject(object: NSManagedObject, changedForPickerTF pickerTF: CDPickerTextField) {
            
        if let item = self.selectedObject as? Item {
            switch (pickerTF) {
                case self.unitPickerTextField:
                    if let unit = object as? Unit {
                        item.unit = unit
                    }
                    break;
                case self.homeLocationPickerTextField:
                    if let locationAtHome = object as? LocationAtHome {
                        item.locationAtHome = locationAtHome
                    }
                    break;
                case self.shopLocationPickerTextField:
                    if let locationAtShop = object as? LocationAtShop {
                        item.locationAtShop = locationAtShop
                    }
                    break;
                default:
                    break;
            }
        } else {self.done()}
        self.refreshInterface()
    }
    
    // MARK: - INTERACTION
    var activeField:UITextField?
    func keyboardDidShow (note:NSNotification) {
           
        if let info = note.userInfo, keyboardFrameEndUserInfoKey = info[UIKeyboardFrameEndUserInfoKey] as? NSValue {
        
            // Find top of keyboard input view (i.e. picker)
            var keyboardRect: CGRect = keyboardFrameEndUserInfoKey.CGRectValue()
            keyboardRect = self.view.convertRect(keyboardRect, fromView: nil)
            let keyboardTop = keyboardRect.origin.y
            
            // Resize scroll view
            var newScrollViewFrame = CGRectMake(0, 0, self.view.bounds.size.width, keyboardTop)
            newScrollViewFrame.size.height = keyboardTop - self.view.bounds.origin.y
            self.scrollView.frame = newScrollViewFrame
            
            // Scroll to the active Text-Field
            if let _activeField = self.activeField {
                self.scrollView.scrollRectToVisible(_activeField.frame, animated: true)
            }    
        }
    }
    func keyboardWillHide () {
        
        // Reset Scrollview to the same size as the containing view
        let defaultFrame = CGRectMake(self.scrollView.frame.origin.x, self.scrollView.frame.origin.y, self.view.frame.size.width, self.view.frame.size.height)
        self.scrollView.frame = defaultFrame
            
        // Scroll to the top again
        let topFrame = CGRectMake(0, 0, 1, 1)
        self.scrollView.scrollRectToVisible(topFrame, animated: true)
    }
 
    // MARK: - CAMERA
    func checkCamera () {
        self.cameraButton.enabled = UIImagePickerController.isSourceTypeAvailable(UIImagePickerControllerSourceType.Camera)
    }
    @IBAction func showCamera () {
            
        if UIImagePickerController.isSourceTypeAvailable(UIImagePickerControllerSourceType.Camera) == true {
        self.camera = UIImagePickerController()

            if let _camera = self.camera {
                    
                _camera.sourceType = UIImagePickerControllerSourceType.Camera
                _camera.mediaTypes = UIImagePickerController.availableMediaTypesForSourceType(UIImagePickerControllerSourceType.Camera)!
                _camera.allowsEditing = true
                _camera.delegate = self
                    
                if let nc = self.navigationController {
                    nc.presentViewController(_camera, animated: true, completion: nil)
                } else {print("Failed to prepare the camera (navigation controller issue)")}
            } else {print("Failed to prepare the camera")}
        } else {print("Camera is not available")}
    }
    func imagePickerController(picker: UIImagePickerController, didFinishPickingImage image: UIImage, editingInfo: [String : AnyObject]?) {
        if let item = self.selectedObject as? Item {
            
            // REMOVE OLD PHOTO
            if let existingPhoto = item.photo {
                CDCloudSync.shared.setCacheStatusForObject(existingPhoto, requestedStatus: DELETING)
            }
            item.photo = nil
            item.thumbnail = nil
            
            // RESIZE PHOTO
            let newsize = CGSizeMake(300, 300)
            UIGraphicsBeginImageContextWithOptions(newsize, true, 1.0)
            image.drawInRect(CGRectMake(0, 0, newsize.width, newsize.height))
            let scaledImage = UIGraphicsGetImageFromCurrentImageContext()
            UIGraphicsEndImageContext()
            
            // ADD NEW PHOTO
            if let photoData = UIImageJPEGRepresentation(scaledImage, 0.7) {
                print("Captured a \(scaledImage.size.height) x \(scaledImage.size.width) photo with size \(photoData.length)")
                if let itemPhoto = NSEntityDescription.insertNewObjectForEntityForName("Item_Photo", inManagedObjectContext: CDHelper.shared.context) as? Item_Photo {
                
                    itemPhoto.modified = NSDate()
                    CDCloudSync.addGroupInfo(itemPhoto)
                    itemPhoto.data = photoData
                    item.photo = itemPhoto

                    CDHelper.saveSharedContext()
                } else {print("\(__FUNCTION__) failed to insert new Item_Photo object")}
            } else {print("\(__FUNCTION__) failed to capture and convert an image")}
        } else {self.done()}
        picker.dismissViewControllerAnimated(true, completion: nil)
    }
    func imagePickerControllerDidCancel(picker: UIImagePickerController) {
        picker.dismissViewControllerAnimated(true, completion: nil)
    }
    
}
