//
//  PrepareTVC.swift
//  Groceries
//
//  Created by Tim Roadley on 30/09/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import CoreData
import CloudKit

class PrepareTVC: CDTableViewController {
    
    // MARK: - CELL CONFIGURATION
    override func configureCell(cell: UITableViewCell, atIndexPath indexPath: NSIndexPath) {
        
        if let item = frc.objectAtIndexPath(indexPath) as? Item {
            
            // Display item photo
            if let imageView = cell.imageView {
                if let thumbnail = item.thumbnail {
                    imageView.image = UIImage(data: thumbnail)
                } else {
                     imageView.image = nil
                }
            }
            
            var itemName = ""
            if let name = item.name {itemName = " \(name)"}
            
            // Prefix quantity and unit
            if let unit = item.unit?.name {itemName = "\(unit)\(itemName)"}
            if let quantity = item.quantity {itemName = "\(quantity)\(itemName)"}
            
            if let textLabel = cell.textLabel, listed = item.listed {
                textLabel.text = itemName
                cell.accessoryType = UITableViewCellAccessoryType.DetailButton
                
                // Color items according to the listed attribute
                if listed.boolValue {
                    textLabel.font = UIFont(name: "Helvetica Neue", size: 18)
                    textLabel.textColor = UIColor(red: 1, green: 0.2, blue: 0.2, alpha: 1)
                } else {
                    textLabel.font = UIFont(name: "Helvetica Neue", size: 16)
                    textLabel.textColor = UIColor.grayColor()
                }
            } else {print("ERROR getting textLabel in \(__FUNCTION__)")}
        } else {print("ERROR getting item in \(__FUNCTION__)")}
    }
    
    // MARK: - INITIALIZATION
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
    
        // CDTableViewController subclass customization
        self.entity = "Item"
        self.sort = [NSSortDescriptor(key: "locationAtHome.storedIn", ascending: true),NSSortDescriptor(key: "name", ascending: true)]
        self.sectionNameKeyPath = "locationAtHome.storedIn"
        self.fetchBatchSize = 25
    
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "groupChanged", name: "GroupChanged", object: nil)
    }
    
    // MARK: - VIEW
    override func viewDidAppear(animated: Bool) {
        super.viewDidAppear(animated)
        Item.ensureLocationsAreNotNilForAllItems()
        self.performFetch()
        
        // Create missing Thumbnails
        let sort = self.frc.fetchRequest.sortDescriptors
        let size = CGSizeMake(66, 66)
        let moc  = CDHelper.shared.importContext
        CDThumbnailer.createMissingThumbnails("Item", thumbnailAttribute: "thumbnail", photoRelationship: "photo", photoAttribute: "data", sort: sort, size: size, moc: moc)
        
        // Trigger Deduplication
        CDDeduplicator.deDuplicateEntityWithName("Item", uniqueAttributeName: "name", backgroundMoc: CDHelper.shared.importContext)
        CDDeduplicator.deDuplicateEntityWithName("Unit", uniqueAttributeName: "name", backgroundMoc: CDHelper.shared.importContext)
        CDDeduplicator.deDuplicateEntityWithName("LocationAtHome", uniqueAttributeName: "storedIn", backgroundMoc: CDHelper.shared.importContext)
        CDDeduplicator.deDuplicateEntityWithName("LocationAtShop", uniqueAttributeName: "aisle",    backgroundMoc: CDHelper.shared.importContext)
        Item.ensureLocationsAreNotNilForAllItems()
    }
    override func viewDidLoad() {
        super.viewDidLoad()
        self.configureSearch()

        // CloudSync
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "syncMessage:", name: "syncMessage", object: nil)
        self.refreshControl = UIRefreshControl()
        if let refreshControl = self.refreshControl {
            refreshControl.addTarget(self, action: Selector("sync"), forControlEvents: UIControlEvents.ValueChanged)
        }
    }
    
    // MARK: - DATA SOURCE: UITableView
    override func tableView(tableView: UITableView, commitEditingStyle editingStyle: UITableViewCellEditingStyle, forRowAtIndexPath indexPath: NSIndexPath) {
        
        if (editingStyle == UITableViewCellEditingStyle.Delete) {
            
            if let object = self.frc.objectAtIndexPath(indexPath) as? NSManagedObject {
                self.frc.managedObjectContext.deleteObject(object)
            }
            CDHelper.saveSharedContext()
        }
    }
    override func tableView(tableView: UITableView, didSelectRowAtIndexPath indexPath: NSIndexPath) {
            
        // Add/Remove items from the shopping list
        if let item = self.frc.objectAtIndexPath(indexPath) as? Item, listed = item.listed {
            if listed.boolValue {
                item.listed = NSNumber(bool: false)
            } else {
                item.listed = NSNumber(bool: true)
                item.collected = NSNumber(bool: false)
            }
            CDCloudSync.shared.setCacheStatusForObject(item, requestedStatus: CHANGED)
            CDHelper.saveSharedContext()
        } 
    }
    
    // MARK: - INTERACTION
    @IBAction func clear (sender: AnyObject) {
        
        if let request = CDHelper.shared.model.fetchRequestTemplateForName("ShoppingList") {
                
            let context = frc.managedObjectContext
            var error: NSError?
        
            let listedItemCount = context.countForFetchRequest(request, error: &error)
            if let _error = error {print("ERROR getting 'ShoppingList' fetch request template: \(_error.localizedDescription)")}
            if listedItemCount > 0 {
      
                // ACTION CONFIRMATION
                let sheet = UIAlertController(title: "Clear Entire Shopping List?", message:nil, preferredStyle:.ActionSheet)
                let confirm =  UIAlertAction(title: "Clear", style: .Destructive, handler: { (action) -> Void in
                    self.clearList()
                })
                let cancel = UIAlertAction(title: "Cancel", style: .Cancel, handler: { (action) -> Void in
                    // do nothing
                })
                sheet.addAction(confirm)
                sheet.addAction(cancel)
                self.presentViewController(sheet, animated: true, completion: nil)
        
            } else {
        
                // ALERT NOTIFICATION
                let alert = UIAlertController(title: "Nothing to Clear", message: "Add items to the Shop tab by tapping them on the Prepare tab. Remove all items from the Shop tab by clicking Clear on the Prepare tab", preferredStyle: .Alert)
                 let ok = UIAlertAction(title: "Ok", style:.Default, handler: { (action) -> Void in
                     // Do nothing
                 })
                 alert.addAction(ok)
                self.presentViewController(alert, animated: true, completion: nil)
            }
        } else {
            print("ERROR getting 'ShoppingList' fetch request")
        }
        CDHelper.saveSharedContext()
    }
    func clearList () {
            
        if let request = CDHelper.shared.model.fetchRequestTemplateForName("ShoppingList") {
        
            let context = frc.managedObjectContext
                
            context.performBlock {
                    
                do {
                    if let items = try context.executeFetchRequest(request) as? [Item] {
                    
                        for item in items {
                            item.listed = NSNumber(bool: false)
                            CDCloudSync.shared.setCacheStatusForObject(item, requestedStatus: CHANGED)
                        }
                    } else {print("\(__FUNCTION__) FAILED to fetch objects")}
                } catch {print("\(__FUNCTION__) FAILED to fetch objects")}
                CDHelper.saveSharedContext()
            }
         } else {print("ERROR getting 'ShoppingList' fetch request")}
    }
    
    // MARK: - SEGUE
    override func prepareForSegue(segue: UIStoryboardSegue, sender: AnyObject?) {

        if let itemVC = segue.destinationViewController as? ItemVC {

            if segue.identifier == "Add Item Segue" {

                let object = NSEntityDescription.insertNewObjectForEntityForName("Item", inManagedObjectContext: CDHelper.shared.context)
                CDCloudSync.addGroupInfo(object)
                itemVC.segueObject = object
            } else {
                print("Unidentified Segue Attempted!")
            }
        }
    }
    override func tableView(tableView: UITableView, accessoryButtonTappedForRowWithIndexPath indexPath: NSIndexPath) {
            
        if let itemVC = self.storyboard?.instantiateViewControllerWithIdentifier("ItemVC") as? ItemVC {
            if let object = self.frc.objectAtIndexPath(indexPath) as? NSManagedObject {
                itemVC.segueObject = object
            }
            self.navigationController?.pushViewController(itemVC, animated: true)
        }
    } 
  
    // MARK: - CLOUDSYNC
    func groupChanged () {
        let predicate = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.filter = predicate
        self.frc.fetchRequest.predicate = predicate
        self.performFetch()
    }
    func sync () {
        let defaultContainer = CKContainer.defaultContainer()
        let publicDatabase = defaultContainer.publicCloudDatabase
        CDCloudSync.shared.syncEntities(["Item","Item_Photo","LocationAtHome","LocationAtShop","Unit"], database: publicDatabase, context: CDHelper.shared.importContext)
    }
    func syncMessage (note:NSNotification) {
        
        if let title = note.object as? String {
            self.navigationItem.title = title
        
            if let refreshControl = self.refreshControl {
                refreshControl.endRefreshing()
                self.performFetch()
            }
        }
    }
} 
