//
//  ShopTVC.swift
//  Groceries
//
//  Created by Tim Roadley on 2/10/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import CoreData
import CloudKit

class ShopTVC: CDTableViewController {
    
    // MARK: - CELL CONFIGURATION
    override func configureCell(cell: UITableViewCell, atIndexPath indexPath: NSIndexPath) {
        
        if let item = frc.objectAtIndexPath(indexPath) as? Item {
            
            // Display item photo
            if let imageView = cell.imageView {
                if let thumbnail = item.thumbnail {
                imageView.image = UIImage(data: thumbnail)
            } else {
                imageView.image = nil
                }
            }
            
            var itemName = ""
            if let name = item.name {itemName = " \(name)"}
            
            // Prefix quantity and unit
            if let unit = item.unit?.name {itemName = "\(unit)\(itemName)"}
            if let quantity = item.quantity {itemName = "\(quantity)\(itemName)"}
    
            if let textLabel = cell.textLabel, collected = item.collected {
                textLabel.text = itemName
                cell.accessoryType = UITableViewCellAccessoryType.DetailButton
                
                // Color items depending to show if they're collected
                if collected.boolValue {
                    cell.accessoryType = UITableViewCellAccessoryType.Checkmark
                    textLabel.font = UIFont(name: "Helvetica Neue", size: 16)
                    textLabel.textColor = UIColor(red:0.37, green:0.74, blue:0.35, alpha:1)
                    textLabel.attributedText = NSAttributedString(string: itemName, attributes: [NSStrikethroughStyleAttributeName : "1"])
            
                } else {
                    cell.accessoryType = UITableViewCellAccessoryType.DetailButton
                    textLabel.font = UIFont(name: "Helvetica Neue", size: 18)
                    textLabel.textColor = UIColor(red: 1, green: 0.2, blue: 0.2, alpha: 1)                
                    textLabel.attributedText = NSAttributedString(string: itemName, attributes: [NSStrikethroughStyleAttributeName : "0"])
                }
            } else {print("ERROR getting textLabel in \(__FUNCTION__)")}
        } else {print("ERROR getting item in \(__FUNCTION__)")}
    }
    
    // MARK: - INITIALIZATION
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
    
        // CDTableViewController subclass customization
        self.entity = "Item"
        self.sort = [NSSortDescriptor(key: "locationAtShop.aisle", ascending: true),NSSortDescriptor(key: "name", ascending: true)]
        self.filter = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i AND listed == 1", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.sectionNameKeyPath = "locationAtShop.aisle"
        self.fetchBatchSize = 25
    
        /*if let template = CDHelper.shared.model.fetchRequestTemplateForName("ShoppingList") {
            let request = template as NSFetchRequest
            self.filter = request.predicate
        } else {print("Could not find a fetch request template called ShoppingList")}*/
        
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "groupChanged", name: "GroupChanged", object: nil)
    }
    
    // MARK: - VIEW
    override func viewDidAppear(animated: Bool) {
        super.viewDidAppear(animated)
        Item.ensureLocationsAreNotNilForAllItems()
        self.performFetch()
        
        // Create missing Thumbnails
        let sort = self.frc.fetchRequest.sortDescriptors
        let size = CGSizeMake(66, 66)
        let moc  = CDHelper.shared.importContext
        CDThumbnailer.createMissingThumbnails("Item", thumbnailAttribute: "thumbnail", photoRelationship: "photo", photoAttribute: "data", sort: sort, size: size, moc: moc)
    }
    override func viewDidLoad() {
        super.viewDidLoad()
        
        // CloudSync
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "syncMessage:", name: "syncMessage", object: nil)
        self.refreshControl = UIRefreshControl()
        if let refreshControl = self.refreshControl {
            refreshControl.addTarget(self, action: Selector("sync"), forControlEvents: UIControlEvents.ValueChanged)
        }
    }
    
    // MARK: - DATA SOURCE: UITableView
    override func tableView(tableView: UITableView, didSelectRowAtIndexPath indexPath: NSIndexPath) {
            
        // Check items off the shopping list
        if let item = self.frc.objectAtIndexPath(indexPath) as? Item, collected = item.collected {
            if collected.boolValue {
                 item.collected = NSNumber(bool: false)
            } else {
                 item.collected = NSNumber(bool: true)
            }
            CDCloudSync.shared.setCacheStatusForObject(item, requestedStatus: CHANGED)
            CDHelper.saveSharedContext()
        }         
    }
    
    // MARK: - INTERACTION
    @IBAction func clear (sender: AnyObject) {
        
        if let items = self.frc.fetchedObjects as? [Item] {
                
            var nothingCleared = true
                
            // Alert if there are no listed items
            if items.count == 0 {
                        
                // ALERT NOTIFICATION
                let alert = UIAlertController(title: "Nothing to Clear", message: "Add items using the Prepare tab", preferredStyle: .Alert)
                let ok = UIAlertAction(title: "Ok", style:.Default, handler: { (action) -> Void in
                    // Do nothing
                })
                alert.addAction(ok)
                self.presentViewController(alert, animated: true, completion: nil)
                return
            }
                
            // Clear collected items
            for item in items {
            
                if let collected = item.collected where collected.boolValue {
                    item.listed = NSNumber(bool: false)
                    item.collected = NSNumber(bool: false)
                    nothingCleared = false
                    CDCloudSync.shared.setCacheStatusForObject(item, requestedStatus: CHANGED)
                }
            }
                
            // Alert if there are no collected items
            if nothingCleared {
                        
                // ALERT NOTIFICATION
                let alert = UIAlertController(title: "Nothing to Clear", message: "Select items to be removed from the list before pressing Clear", preferredStyle: .Alert)
                let ok = UIAlertAction(title: "Ok", style:.Default, handler: { (action) -> Void in
                    // Do nothing
                })
                alert.addAction(ok)
                self.presentViewController(alert, animated: true, completion: nil)
            }
        }
        CDHelper.saveSharedContext()
    }
    
    // MARK: - SEGUE
    override func tableView(tableView: UITableView, accessoryButtonTappedForRowWithIndexPath indexPath: NSIndexPath) {
            
        if let itemVC = self.storyboard?.instantiateViewControllerWithIdentifier("ItemVC") as? ItemVC {
            if let object = self.frc.objectAtIndexPath(indexPath) as? NSManagedObject {
                itemVC.segueObject = object
            }
            self.navigationController?.pushViewController(itemVC, animated: true)
        }
    }
    
    // MARK: - CLOUDSYNC
    func groupChanged () {
        let predicate = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i AND listed == 1", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.filter = predicate
        self.frc.fetchRequest.predicate = predicate
        self.performFetch()
    }
    func sync () {
        let defaultContainer = CKContainer.defaultContainer()
        let publicDatabase = defaultContainer.publicCloudDatabase
        CDCloudSync.shared.syncEntities(["Item","Item_Photo","LocationAtHome","LocationAtShop","Unit"], database: publicDatabase, context: CDHelper.shared.importContext)
    }
    func syncMessage (note:NSNotification) {
        
        if let title = note.object as? String {
            self.navigationItem.title = title
        
            if let refreshControl = self.refreshControl {
                refreshControl.endRefreshing()
                self.performFetch()
            }
        }
    }
    
    // MARK: - FAVORITE
    @IBAction func addFavorites () {

        let sheet = UIAlertController(title:"Add Favorite Items to the List?", message:nil, preferredStyle:.ActionSheet)
        
        // ADD BUTTON
        let add = UIAlertAction(title: "Add", style: .Destructive) { (action) -> Void in
            
            let filter = NSPredicate(format: "favorite == 1 AND listed == 0")
            
            if let favoriteItems = CDOperation.objectsForEntity("Item", context: CDHelper.shared.context, filter: filter, sort: nil) as? [Item] {
                
                for item in favoriteItems {
                
                    item.listed = NSNumber(bool: true)
                    item.collected = NSNumber(bool: false)
                    item.modified = NSDate()
                    CDCloudSync.shared.setCacheStatusForObject(item, requestedStatus: CHANGED)
                }
                CDHelper.saveSharedContext()
            }
        }
        sheet.addAction(add)
        
        // CANCEL BUTTON
        let cancel = UIAlertAction(title: "Cancel", style: UIAlertActionStyle.Cancel) { (action) -> Void in
            // Do nothing
        }
        sheet.addAction(cancel)
        
        self.presentViewController(sheet, animated: true, completion: nil)
    }
}
